use crate::usernames::validators::{validate_pubkey, validate_relays};
use crate::utils::error::Error;
use nostr::key::XOnlyPublicKey;
use nostr::prelude::*;
use regex::Regex;
use serde::{Deserialize, Serialize};
use validator::Validate;

lazy_static! {
    static ref VALID_CHARACTERS: Regex = Regex::new(r"^[a-zA-Z0-9\_]+$").unwrap();
}

#[derive(Serialize, Deserialize, Debug, Validate, PartialEq, Clone)]
pub struct UserBody {
    #[validate(length(min = 1), regex = "VALID_CHARACTERS")]
    pub name: String,
    #[validate(custom(function = "validate_pubkey"))]
    pubkey: String,
    #[validate(custom(function = "validate_relays"))]
    pub relays: Vec<String>
}

impl UserBody {
    pub fn get_pubkey(&self) -> String {
        nostr::Keys::from_pk_str(&self.pubkey).unwrap().public_key().to_string()
    }
}

#[derive(Serialize, Deserialize, Debug, Validate, Clone)]
pub struct UserQuery {
    #[validate(length(min = 1))]
    pub name: String,
}

#[derive(Serialize, Deserialize, Debug, Clone, Validate)]
pub struct Account {
    #[validate(custom(function = "validate_pubkey"))]
    pub pubkey: String,
    #[validate(length(min = 1), regex = "VALID_CHARACTERS")]
    pub name: String,
    #[validate(length(min = 1))]
    pub password: String,
}

impl Account {
    pub fn get_pubkey(&self) -> Result<XOnlyPublicKey, Error> {
        match nostr::Keys::from_pk_str(&self.pubkey) {
            Ok(pk) => Ok(pk.public_key()),
            Err(e) => Err(Error::invalid_param("pubkey", self.pubkey.clone())),
        }
    }
}

#[derive(Serialize, Deserialize, Debug, Clone, Validate)]
pub struct AccountPubkey {
    #[validate(custom(function = "validate_pubkey"))]
    pub pubkey: String,
}
