use serde::{Deserialize, Serialize};
use serde_json;
use std::error::Error as StdError;
use std::{
    convert::From,
    fmt::{self, Display},
};
use validator::ValidationErrors;
use warp::{http::StatusCode, reject::Reject};

const VERSION: &str = env!("CARGO_PKG_VERSION");

#[derive(Clone, Debug, Deserialize, Serialize, PartialEq)]
pub struct Error {
    pub code: u16,
    pub message: String,
    pub sneedstr_version: String,
}

impl StdError for Error {
    fn source(&self) -> Option<&(dyn StdError + 'static)> {
        None
    }

    fn cause(&self) -> Option<&dyn StdError> {
        self.source()
    }
}

impl Error {
    pub fn new(code: StatusCode, message: String) -> Self {
        Self {
            code: code.as_u16(),
            message,
            sneedstr_version: VERSION.to_string(),
        }
    }

    pub fn from_anyhow_error(code: StatusCode, err: anyhow::Error) -> Self {
        Self::new(code, err.to_string())
    }

    pub fn bad_request<S: Display>(msg: S) -> Self {
        Self::new(StatusCode::BAD_REQUEST, msg.to_string())
    }

    pub fn internal_with_message<S: Display>(msg: S) -> Self {
        Self::new(StatusCode::INTERNAL_SERVER_ERROR, msg.to_string())
    }

    pub fn validation_error(msg: ValidationErrors) -> Self {
        let message =
            serde_json::to_string(&msg.field_errors()).unwrap_or("Validation Error".to_string());
        Self::new(StatusCode::BAD_REQUEST, message)
    }

    pub fn not_found<S: Display>(resource: &str, identifier: S) -> Self {
        Self::new(
            StatusCode::NOT_FOUND,
            format!("{} not found by {}", resource, identifier),
        )
    }

    pub fn invalid_param<S: Display>(name: &str, value: S) -> Self {
        Self::bad_request(format!("invalid parameter {}: {}", name, value))
    }

    pub fn invalid_request_body<S: Display>(msg: S) -> Self {
        Self::bad_request(format!("invalid request body: {}", msg))
    }

    pub fn internal(err: anyhow::Error) -> Self {
        Self::from_anyhow_error(StatusCode::INTERNAL_SERVER_ERROR, err)
    }

    pub fn status_code(&self) -> StatusCode {
        StatusCode::from_u16(self.code).unwrap_or(StatusCode::INTERNAL_SERVER_ERROR)
    }
}

impl fmt::Display for Error {
    fn fmt(&self, f: &mut fmt::Formatter<'_>) -> fmt::Result {
        write!(
            f,
            "Error {{ code: {}, message: '{}', sneedstr_version: {:?} }}",
            self.code, self.message, self.sneedstr_version
        )
    }
}

impl Reject for Error {}

impl From<anyhow::Error> for Error {
    fn from(e: anyhow::Error) -> Self {
        Self::internal(e)
    }
}

impl From<serde_json::error::Error> for Error {
    fn from(err: serde_json::error::Error) -> Self {
        Self::internal(err.into())
    }
}

#[cfg(test)]
mod tests {
    use super::Error;
    use warp::http::StatusCode;

    #[test]
    fn test_to_string() {
        let err = Error::new(StatusCode::BAD_REQUEST, "invalid address".to_owned());
        assert_eq!(
            err.to_string(),
            "Error { code: 400, message: 'invalid address', sneedstr_version: \"0.1.1\" }"
        )
    }

    #[test]
    fn test_from_anyhow_error_as_internal_error() {
        let err = Error::from(anyhow::format_err!("hello"));
        assert_eq!(
            err.to_string(),
            "Error { code: 500, message: 'hello', sneedstr_version: \"0.1.1\" }"
        )
    }

    #[test]
    fn test_to_string_with_sneedstr_version() {
        let err = Error::new(StatusCode::BAD_REQUEST, "invalid address".to_owned());
        assert_eq!(
            err.to_string(),
            "Error { code: 400, message: 'invalid address', sneedstr_version: \"0.1.1\" }"
        )
    }

    #[test]
    fn test_internal_error() {
        let err = Error::internal(anyhow::format_err!("hello"));
        assert_eq!(
            err.to_string(),
            "Error { code: 500, message: 'hello', sneedstr_version: \"0.1.1\" }"
        )
    }
}
